import argparse
from pathlib import Path
from itertools import chain
from multiprocessing import Pool
import pickle

import pandas as pd
from sklearn.metrics import (confusion_matrix, ConfusionMatrixDisplay,
                             f1_score,  accuracy_score, precision_score, recall_score)
import matplotlib.pyplot as plt
import seaborn as sns
import numpy as np

SPECIES_TO_ID = {
    'acinetobacter baumannii': 'NZ_CP053098',
    # Two names for the same species here
    'actinomyces odontolyticus': 'NZ_DS26458',
    'schaalia odontolytica': 'NZ_DS26458',
    'bacillus cereus': 'AE017194',
    'bacteroides vulgatus': 'CP000139',
    'clostridium beijerinckii': 'CP000721',
    'deinococcus radiodurans': 'AE000513',
    'enterococcus faecalis': 'CP025020',
    'escherichia coli': 'U00096',
    'helicobacter pylori': 'AE000511',
    'lactobacillus gasseri': 'CP000413',
    'listeria monocytogenes': 'AL591824',
    'neisseria meningitides': 'AE002098',
    'porphyromonas gingivalis': 'AP009380',
    # Never found in ground truth but included for completion
    'propionibacterium acnes': 'AE017283',
    'pseudomonas aeruginosa': 'AE004091',
    'rhodobacter sphaeroides': 'CP00014',
    'staphylococcus aureus': 'CP000730',
    'staphylococcus epidermidis': 'AE015929',
    'streptococcus agalactiae': 'AE009948',
    'streptococcus mutans': 'AE014133',
    'streptococcus pneumoniae': 'AE005672',
}


def make_balanced_df(df):
    """From dataframe, make sure each species is present 100 times

    :param df: dataframe that contains column called 'species id'
    :type df: pd.Dataframe
    :return: dataframe with balanced species
    """
    # Remove species that are too rare
    df = df.groupby('species id').filter(lambda x: len(x) > 100)
    df = df.groupby('species id').sample(100)
    return df


def get_performance_metrics_from_predictions(y_pred, y_true):
    """Given two arrays of predict and true labels, return accuracy score,
    confusion matrix and f1 score"""
    f1 = f1_score(y_true, y_pred, zero_division=0)
    accuracy = accuracy_score(y_true, y_pred)
    cm = confusion_matrix(y_true, y_pred)
    return accuracy, cm, f1


def parse_sequencing_summary(file_path, ground_truth, return_df=False):
    """Parse a sequencing_summary.txt output by guppy

    :param file_path: Path to sequencing_summary.txt. Should be in the form of
    **/guppy/{target}/fold{ix}/sequencing_summary.txt
    :type file_path: Path
    :param ground_truth: Dataframe that contains ground truth, can be read from
     csv that is output by tools.set_ground_truths_of_reads.py
    :type ground_truth: pd.DataFrame
    :return: tuple with tool, target_species, fold, accuracy,
    f1 and confusion matrix
    """
    print(f'Processing {file_path}')
    # Extract some properties from the file path
    fold = file_path.parts[-2][-1]
    target_species = file_path.parts[-3].replace('_', ' ')
    target_id = SPECIES_TO_ID[target_species]
    tool = file_path.parts[-4]
    assert tool == 'guppy'

    df_guppy = pd.read_csv(file_path, sep='\t')
    df_guppy.rename(columns={'read_id': 'read id'}, inplace=True)
    df = df_guppy.merge(ground_truth, on='read id')
    df = make_balanced_df(df)
    # print('Guppy accuracy on df of size', df.shape)

    y_pred = df['alignment_identity'].apply(lambda x: 0 if x < 0.95 else 1)
    df['y_pred'] = y_pred

    # Set truth to 1 if ground truth species is found
    y_true = df['species id'].apply(lambda x:
                                    1 if x.find(target_id) == 0
                                    else 0)
    df['y_true'] = y_true
    if return_df:
        return df
    accuracy, cm, f1 = get_performance_metrics_from_predictions(y_pred, y_true)
    return tool, target_species, fold, accuracy, f1, cm


def parse_paf(paf_path, ground_truth):
    """Parse single paf output of uncalled or minimap2

    :param paf_path: path to .paf file output. Should be mapped
    to single species reference genome.
    If mapped to no species, the target sequence name should be '*'
    :param ground_truth: Dataframe that contains ground truth, can be read from
     csv that is output by tools.set_ground_truths_of_reads.py
    :type ground_truth: pd.DataFrame
    :return: tuple with tool, target_species, fold, accuracy, f1
    """
    headers = [
        "read id",
        "Query sequence length",
        "Query start (0-based; BED-like; closed)",
        "Query end (0-based; BED-like; open)",
        "Relative strand: + or -",
        "Target sequence name",
        "Target sequence length",
        "Target start on original strand (0-based)",
        "Target end on original strand (0-based)",
        "Number of residue matches",
        "Alignment block length",
        "Mapping quality (0-255; 255 for missing)"
    ]

    print(f'Processing {paf_path}')
    fold = paf_path.stem[-1]
    target_species = paf_path.parts[-2].replace('_', ' ')
    target_id = SPECIES_TO_ID[target_species]
    tool = paf_path.parts[-3]

    uncalled_paf = pd.read_csv(paf_path, sep='\t', names=headers,
                               usecols=range(12))
    df = uncalled_paf.merge(ground_truth, on='read id')
    df = make_balanced_df(df)
    # print(f'{tool} accuracy on df of size', df.shape)
    # Set truth to 1 if ground truth species is found
    y_true = df['species id'].apply(lambda x:
                                           1 if x.find(target_id) == 0
                                           else 0)
    if tool == 'deepnano':
        query_len = np.array(df["Query sequence length"])
        alignment_length = np.array(df["Alignment block length"])
        align_perc = alignment_length / query_len
        y_pred = (align_perc > 0.90).astype(int)
    elif tool == 'uncalled':
        # Mapping to a * means mapping to no species
        y_pred = df['Target sequence name'].apply(
            lambda x: 0 if x == '*' else 1)
    else:
        raise ValueError('Tool not found')
    accuracy, cm, f1 = get_performance_metrics_from_predictions(y_pred, y_true)
    return tool, target_species, fold, accuracy, f1, cm


def parse_squigglenet_output(in_dir, ground_truth):
    """Parse directory with inference txt files output by squigglenet

    :param in_dir: path to directory with batch_?.txt
    :type in_dir: Path
    :param ground_truth: dataframe of ground truth
    :type ground_truth: pd.Dataframe
    :return: tuple with tool, target_species, fold, accuracy, f1
    """
    print(f'Processing {in_dir}')
    fold = in_dir.parts[-2][-1]
    target_species = in_dir.parts[-3].replace('_', ' ')
    target_id = SPECIES_TO_ID[target_species]
    tool = in_dir.parts[-4]
    assert tool == 'squigglenet'

    predictions = []
    for file in in_dir.iterdir():
        with file.open() as f:
            temp = f.readlines()
            temp = [i.split() for i in temp]
            predictions.extend(temp)

    df = pd.DataFrame.from_records(predictions, columns=['read id', 'y pred'])

    merged_df = df.merge(ground_truth, on='read id')
    merged_df = make_balanced_df(merged_df)
    # print('Squigglenet accuracy on df of size', merged_df.shape)
    y_true = merged_df['species id'].apply(lambda x:
                                           1 if x.find(target_id) == 0
                                           else 0)
    y_pred = merged_df['y pred'].astype(int)
    # Flip predictions because squigglenet be like that
    y_flipped = 1 - y_pred
    accuracy, cm, f1 = get_performance_metrics_from_predictions(y_flipped,
                                                                y_true)
    return tool, target_species, fold, accuracy, f1, cm


def parse_baseless_output(file, ground_truth, for_pr_plot=False):
    """

    :param file: Path to baseless output file
    :type file: Path
    :param ground_truth:
    :return:
    """
    print(f'Processing {file}')
    fold = file.stem[-1]
    target_species = file.parts[-2].replace('_', ' ')
    target_id = SPECIES_TO_ID[target_species]
    tool = file.parts[-3]

    baseless_output = pd.read_csv(file, names=['file name', 'species_is_found'])
    merged_df = baseless_output.merge(ground_truth, on='file name')
    # print('Baseless accuracy on df of size', merged_df.shape)
    y_true = merged_df['species id'].apply(lambda x:
                                           1 if x.find(target_id) == 0
                                           else 0)
    y_pred = [1 if float(i) > 0.55 else 0 for i in merged_df.species_is_found]
    if not y_pred:
        # No positive examples
        return None
    if for_pr_plot:
        # Return precision and recall, used when plotting baseLess'
        # precision and recall per species
        print(target_species, fold)
        precision = precision_score(y_true, y_pred)
        recall = recall_score(y_true, y_pred)
        return target_species, fold, precision, recall
    accuracy, cm, f1 = get_performance_metrics_from_predictions(y_pred, y_true)
    return tool, target_species, fold, accuracy, f1, cm


def calculate_accuracy_from_output(args):
    """Function that can be called from main. Takes raw tool output data and
    calculates accuracy, f1 score and confusion matrix for each tool, species
    and fold. Returns this data as a dataframe and also saves it"""
    ground_truth = pd.read_csv(args.ground_truth)
    # Find accuracy files
    assert args.benchmark_path.exists(), 'Benchmark directory not found'
    baseless_files = args.benchmark_path.glob(
        "baseless_8_and_9mers/*/output_fold?.csv")
    deepnano_files = args.benchmark_path.glob(
        "deepnano/*/minimap2_deepnano_fold?.paf")
    uncalled_files = args.benchmark_path.glob(
        "uncalled/*/uncalled_out_fold?.paf")
    guppy_files = args.benchmark_path.glob(
        "guppy/*/fold?/sequencing_summary.txt")
    squigglenet_files = args.benchmark_path.glob(
        "squigglenet/*/fold?/inference")
    # Prepare arguments to parse to the functions in the pool
    args_baseless = [[file, ground_truth] for file in baseless_files]
    args_uncalled = [[file, ground_truth] for file in uncalled_files]
    args_guppy = [[file, ground_truth] for file in guppy_files]
    args_squigglenet = [[folder, ground_truth]
                        for folder in squigglenet_files]
    with Pool(args.threads) as p:
        baseless_results = p.starmap(parse_baseless_output, args_baseless)
        squigglenet_results = p.starmap(parse_squigglenet_output,
                                        args_squigglenet)
        uncalled_results = p.starmap(parse_paf, args_uncalled)
        guppy_results = p.starmap(parse_sequencing_summary, args_guppy)
    deepnano_results = [parse_paf(file, ground_truth) for file in deepnano_files]
    all_records = (deepnano_results + uncalled_results + guppy_results
                   + squigglenet_results + baseless_results)
    assert all_records, ('No files found: make sure the input folder '
                         'contains directories called "uncalled",'
                         ' "guppy", "baseless" and "deepnano"')
    df = pd.DataFrame.from_records(all_records,
                                   columns=['tool', 'species',
                                            'fold', 'accuracy', 'f1',
                                            'confusion matrix'])
    # # Create confusion matrices
    # cm_series = df.groupby(['tool', 'species'])['confusion matrix'].aggregate(
    #     np.sum)
    # args.out_dir.mkdir(parents=True, exist_ok=True)
    # with open(args.out_dir / f'confusion_matrices.pickle', 'wb') as f:
    #     pickle.dump(cm_series, f)
    #
    # # Confusion matrix out directory
    # cm_out_dir = args.out_dir / 'confusion_matrices'
    # cm_out_dir.mkdir(exist_ok=True)
    # for (tool, species), cm in cm_series.iteritems():
    #     ConfusionMatrixDisplay(cm).plot()
    #     plt.savefig(cm_out_dir / f'{tool}_{species}.png')
    Path(args.out_dir).mkdir(exist_ok=True)
    df.to_csv(args.out_dir / f'all_algorithms_performance.csv', index=False)
    return df


def main(args):
    # For debugging
    pd.options.display.width = 0

    if not args.input_performance_csv:
        df = calculate_accuracy_from_output(args)
    else:
        df = pd.read_csv(args.input_performance_csv)

    # Remove gingivalis from dataset
    is_not_gingivalis = df['species'].apply(lambda x: False
                                            if x.find('gingivalis') > 0
                                            else True)
    df = df[is_not_gingivalis]
    # Temporary; remove A. odontolyliticus fold 1 because it contained no pos examples
    i_to_remove = df[((df.tool == 'baseless')
                      & (df.species =='actinomyces odontolyticus')
                      & (df.fold == 1))].index
    df.drop(i_to_remove, inplace=True)
    # Set species in alphabetical order
    all_species = df.species.unique()
    all_species.sort()

    g = sns.catplot(x='species', y='accuracy', ci='sd', data=df, hue='tool',
                    kind='bar', order=all_species, legend_out=False, aspect=2)
    g.set_xticklabels(rotation=45, ha="right",
                      rotation_mode="anchor")
    # g.ax.set_title('5-fold CV benchmarked model accuracy')
    # g.fig.suptitle('5-fold CV model accuracy per bacterial species')
    g.tight_layout()
    g.savefig(args.out_dir / 'all_algorithms_performance_accuracy.svg')

    g = sns.catplot(x='species', y='f1', ci='sd', data=df, hue='tool',
                    kind='bar', order=all_species, legend=False, aspect=2)
    g.set_xticklabels(rotation=45, ha="right",
                      rotation_mode="anchor")
    # g.ax.set_title(r'5-fold CV model $F_1$ score per bacterial species')
    g.tight_layout()
    g.savefig(args.out_dir / 'all_algorithms_performance_f1.svg')


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description="""Plot accuracy of
        benchmarked algorithms. Provide only performance CSV and out-dir
        or all arguments except for input-performance-csv""")
    parser.add_argument('--benchmark-path',
                        help='Path to folder that contains all benchmarks '
                             'output by the snakemake workflow. '
                             'Should contain a folders called "uncalled",'
                             ' "guppy" and "deepnano"',
                        required=True, type=Path)
    parser.add_argument('--ground-truth',
                        help='Path to csv with ground truth labels. '
                             'It is output by set_ground_truths_of_reads.py',
                        required=True, type=Path)
    parser.add_argument('--input-performance-csv',
                        help='Path to csv that contains model performance as '
                             'output by earlier run of this script. '
                             'If provided, the script will not recalculate '
                             'all performance statistics manually but '
                             'read them from this csv.',
                        required=False, type=Path)
    parser.add_argument('--out-dir',
                        help='Directory in which to save the figures and'
                             ' csv with model performance',
                        required=True, type=Path)
    parser.add_argument('--threads',
                        help='Number of threads to run in parallel',
                        required=False, type=int, default=30)
    args = parser.parse_args()
    main(args)
