from pathlib import Path
import argparse
from compare_benchmark_performance.compare_accuracy_per_read import SPECIES_TO_ID, parse_sequencing_summary
import pandas as pd
import random
from sklearn.metrics import precision_recall_curve, PrecisionRecallDisplay
import matplotlib.pyplot as plt
import numpy as np


def create_in_silico_samples(ground_truth: pd.DataFrame,
                             input_fold_folders: Path, fold_index: int,
                             for_training: bool, target_species,
                             nr_total_examples: int,
                             nr_pos_examples: int) -> list:
    """Create random in silico sample with given amount of target and
    non-target reads for a certain fold and target

    :param ground_truth: ground truth dataframe
    :param input_fold_folders: folder that contains all input folds
    :param fold_index: fold to evaluate
    :param for_training: if True, only look at the train reads for that fold
                         if False, just use the test reads of that fold
    :param target_species: name of target species
    :param nr_total_examples: how big the total sample should be
    :param nr_pos_examples: how many target reads the sample should be
    :return: list with all file names in the given sample
    """
    if for_training:
        # Only look at files outside the test set
        valid_seq_file = input_fold_folders / f'fold{fold_index}' / 'train_reads.txt'
        valid_seqs = valid_seq_file.read_text().splitlines()
    else:
        # Only look at files in test set
        test_dir = input_fold_folders / f'fold{fold_index}' / 'test'
        valid_seqs = [f.name for f in test_dir.iterdir()]
    # Keep only the valid seqs
    df = ground_truth[ground_truth['file name'].isin(valid_seqs)]
    target_id = SPECIES_TO_ID[target_species]
    row_is_target = df['species id'].apply(lambda x:
                                           True if x.find(target_id) == 0
                                           else False)
    target_df = df[row_is_target]
    non_target_df = df[~row_is_target]
    pos_examples = target_df.sample(nr_pos_examples)
    neg_examples = non_target_df.sample(nr_total_examples - nr_pos_examples)

    all_examples = list(pos_examples['file name']) + list(neg_examples['file name'])
    random.shuffle(all_examples)
    return all_examples


def get_all_guppy_output(gt, target):
    """Get dataframe with guppy predictions of each read"""
    guppy_path = Path(
        '/home/noord087/lustre_link/mscthesis/benchmarking/output_16s_files_3_cores/guppy')
    target_dirname = target.replace(' ', '_')
    all_output_paths = (guppy_path / f'{target_dirname}').glob(
        'fold?/sequencing_summary.txt')
    all_output = []
    for output in all_output_paths:
        all_output.append(parse_sequencing_summary(output, gt, return_df=True))
    df = pd.concat(all_output, ignore_index=True)
    return df


def determine_read_count_cutoff_guppy(gt, input_folders, fold, target):
    df = get_all_guppy_output(gt, target)

    nr_total_examples = 5000
    guppy_predictions = []
    ground_truth = []

    for i in range(20):
        print('Iteration', i)
        # Do 20 negative examples
        sample = create_in_silico_samples(gt, input_folders,
                                          fold_index=fold,
                                          for_training=True,
                                          target_species=target,
                                          nr_total_examples=nr_total_examples,
                                          nr_pos_examples=0)
        selected_df = df[df['filename'].isin(sample)]
        frac_guppy = sum(selected_df.y_pred) / nr_total_examples
        guppy_predictions.append(frac_guppy)
        ground_truth.append(0)
        # Do 20 positive examples
        sample = create_in_silico_samples(gt, input_folders,
                                          fold_index=fold,
                                          for_training=True,
                                          target_species=target,
                                          nr_total_examples=nr_total_examples,
                                          nr_pos_examples=50)
        selected_df = df[df['filename'].isin(sample)]
        frac_guppy = sum(selected_df.y_pred) / nr_total_examples
        guppy_predictions.append(frac_guppy)
        ground_truth.append(1)

    # PrecisionRecallDisplay.from_predictions(ground_truth, guppy_predictions)
    # plt.show()
    precision, recall, thresholds = precision_recall_curve(ground_truth, guppy_predictions)
    if not np.any(precision <= 0.99):
        # All thresholds work
        return thresholds[-1]
    best_index = np.argwhere(precision <= 0.99)[0] - 1
    best_treshold = thresholds[best_index]
    return best_treshold


def predict_for_guppy(ground_truth, target, input_folders):
    df = get_all_guppy_output(ground_truth, target)
    # for fold in range(5):
    fold = 0
    threshold = determine_read_count_cutoff_guppy(ground_truth, input_folders, fold, target)

    test_reads = create_in_silico_samples(ground_truth, input_folders, fold,
                                          for_training=False,
                                          target_species=target,
                                          nr_total_examples=5000,
                                          nr_pos_examples=50)
    selected_df = df[df['filename'].isin(test_reads)]
    frac_guppy = sum(selected_df.y_pred) / len(selected_df)
    if frac_guppy > threshold:
        print('True')
    else:
        print('False')


def main(args):
    target = 'escherichia coli'
    ground_truth = pd.read_csv(args.ground_truth)
    determine_read_count_cutoff_guppy(ground_truth, args.input_fold_folders, 0,
                                      target)



if __name__ == '__main__':
    parser = argparse.ArgumentParser(description="""Plot accuracy of
        benchmarked algorithms per in silico sample. 
        Provide only performance CSV and out-dir or all arguments except 
        for input-performance-csv""")
    parser.add_argument('--benchmark-path',
                        help='Path to folder that contains all benchmarks '
                             'output by the snakemake workflow. '
                             'Should contain a folders called "uncalled",'
                             ' "guppy" and "deepnano"', # Todo double check if this docstring is correct
                        required=True, type=Path)
    parser.add_argument('--ground-truth',
                        help='Path to csv with ground truth labels. '
                             'It is output by set_ground_truths_of_reads.py',
                        required=True, type=Path)
    parser.add_argument('--input-fold-folders',
                        help='Path to folders of input folds',
                        required=True, type=Path)
    args = parser.parse_args()
    main(args)
