import tarfile, os, re
import json
import pandas as pd
from shutil import copyfile
from pathlib import Path
from tempfile import TemporaryDirectory
from run_production_pipeline import main as rpp
import subprocess
import warnings
from copy import copy
import yaml
import tensorflow as tf
import tensorflow.keras.backend as K
import h5py
import numpy as np

__location__ = os.path.realpath(os.path.join(os.getcwd(), os.path.dirname(__file__)))

COMPLEMENT_DICT = {'A': 'T', 'T': 'A', 'C': 'G', 'G': 'C'}
kmer_freqs_df = pd.read_parquet(Path(f'{__location__}/../data/ncbi_16S_bacteria_archaea_kmer_counts.parquet').resolve())
kmer_freqs_dict = dict(kmer_freqs_df.sum(axis=0))
# TODO hardcoded this filepath
with open(f'/lustre/BIF/nobackup/noord087/mscthesis/9mer_library_excluding_10_percent/1500_9mers_final.txt', 'r') as fh:
    KMER_GOOD_CANDIDATE_LIST = [km.strip() for km in fh.readlines() if len(km.strip())]

def reverse_complement(km):
    return ''.join([COMPLEMENT_DICT[k] for k in km][::-1])

def fa2kmers(fa_fn, kmer_sizes, available_models=None):
    """
    take multifasta file, return dict of lists of k-mers present in each sequence
    """
    with open(fa_fn, 'r') as fh:
        targets_txt = fh.read()
    targets_list = ['>' + t for t in targets_txt.split('>') if t]
    targets_dict = {re.search('(?<=>)[^\n]+', t).group(0): t for t in targets_list}
    kmer_dict = {}
    with TemporaryDirectory() as tdo:
        for td in targets_dict:
            kmer_dict[td] = []
            with open(tdo + '/target.fasta', 'w') as fh:
                fh.write(targets_dict[td])
            for kmer_size in kmer_sizes:
                subprocess.run(
                    f'jellyfish count -m {kmer_size} -s {4 ** kmer_size} -C -o {tdo}/mer_counts.jf  {fa_fn}',
                    shell=True)
                kmer_dump = subprocess.run(f'jellyfish dump -c {tdo}/mer_counts.jf', shell=True, capture_output=True)
                kmer_list = [km.split(' ')[0] for km in kmer_dump.stdout.decode('utf-8').split('\n')]
                kmer_list = [km for km in kmer_list if len(km) == kmer_size and km in available_models]  # retain k-mers found to score well in discernability
                kmer_list_curated = []  # remove reverse complements
                for km in kmer_list:
                    if reverse_complement(km) in kmer_list_curated: continue
                    kmer_list_curated.append(km)
                kmer_dict[td].extend(kmer_list_curated)
    return kmer_dict


def get_kmer_candidates_16S(kmer_candidates_dict, min_nb_kmers=5, threshold=0.0001, filter_list=None):
    """
    get smallest set of 16S k-mers for which FDR<threshold
    """
    kmer_candidates_selection_dict = {}
    for kc in kmer_candidates_dict:
        p = 1.0
        kmer_list = sorted(kmer_candidates_dict[kc], key=kmer_freqs_dict.get, reverse=True)  # sort by frequency in 16S genomes
        kmer_candidates_selection_dict[kc] = []
        sub_df = kmer_freqs_df.copy()
        sub_freqs_dict = copy(kmer_freqs_dict)
        cc=0
        while cc <= len(kmer_list) and (p >= threshold or len(kmer_candidates_selection_dict[kc]) < min_nb_kmers):
            km = kmer_list.pop()  # pop least frequent k-mer
            kmer_candidates_selection_dict[kc].append(km)

            # Select entries positive for k-mer or its reverse-complement
            km_rc = reverse_complement(km)
            hit_found = False
            if not len(sub_df):  # avoid making new queries if number of accessions is already 0, but min number of models has not been reached
                hit_found=True
            elif km in sub_df.columns and km_rc in sub_df.columns:
                sub_df = sub_df.query(f'{km} or {km_rc}').copy()
                hit_found=True
            elif km in sub_df.columns:
                sub_df.query(km, inplace=True)
                hit_found = True
            elif km_rc in sub_df.columns:
                sub_df.query(km_rc, inplace=True)
                hit_found = True
            if hit_found:
                hits = len(sub_df)
            else:
                hits = 1
            hits = max(hits, 1)
            p *= (hits - 1) / hits  # update FDR
            kmer_list.sort(key=sub_freqs_dict.get, reverse=True)  # re-sort kmer list
            sub_freqs_dict = dict(sub_df.sum(axis=0))
            cc += 1
        if filter_list:
            kmer_candidates_selection_dict[kc] = [km for km in kmer_candidates_selection_dict[kc] if km in filter_list]
    kc_count_dict = {kc: len(kmer_candidates_selection_dict[kc]) for kc in kmer_candidates_selection_dict}
    selected_id = min(kc_count_dict, key=kc_count_dict.get)
    return kmer_candidates_selection_dict[selected_id]


def compile_model(kmer_dict, filter_width, filter_stride, threshold, k_frac):
    k_threshold = round(len(kmer_dict) * k_frac)

    input = tf.keras.Input(shape=(None, 1), ragged=True)
    input_strided = tf.signal.frame(input.to_tensor(default_value=np.nan), frame_length=filter_width, frame_step=filter_stride, axis=1)
    input_strided = tf.keras.layers.Masking(mask_value=np.nan)(input_strided)
    ht_list = []
    for km in kmer_dict:
        mod = tf.keras.models.load_model(f'{kmer_dict[km]}/nn.h5',compile=False)
        mod._name = km
        h = tf.keras.layers.TimeDistributed(mod)(input_strided)
        h = K.max(h, axis=1)
        ht_list.append(K.cast_to_floatx(K.greater(h, threshold)))
    nb_hits = tf.math.reduce_sum(tf.keras.layers.concatenate(ht_list), axis=1)
    output = nb_hits / len(kmer_dict)
    meta_mod = tf.keras.Model(inputs=input, outputs=output)
    meta_mod.compile()
    return meta_mod


def main(args):
    # List for which k-mers models are available
    if args.nn_directory:
        nn_directory = args.nn_directory
    elif args.target_16S:
        nn_directory = f'{__location__}/../16S_db/'
    else:
        raise ValueError('Either provide --nn-directory or --target-16S')
    # List models for which k-mer is available
    available_mod_kmers = {pth.name: str(pth) for pth in Path(nn_directory).iterdir() if pth.is_dir()}
    # available_mod_kmers = {model_pth.parent.name: str(model_pth.parent) for model_pth in expandpath(nn_directory)}
    # Parse target k-mers
    if args.kmer_list:
        with open(args.kmer_list, 'r') as fh:
            kmers = [km.strip() for km in fh.readlines()]
        requested_kmers = [km for km in kmers if len(km)]
        kmers = [km for km in requested_kmers if km in available_mod_kmers]
    elif args.target_16S:
        kmer_sizes = [8, 9]
        kmer_candidates_dict = fa2kmers(args.target_16S, kmer_sizes, available_mod_kmers)  # Collect k-mers per sequence in target fasta
        if args.train_required:
            kmers = get_kmer_candidates_16S(kmer_candidates_dict, args.min_nb_models, 0.0001)
            kmers_no_models = [km for km in kmers if km not in available_mod_kmers]
            if len(kmers_no_models):  # train additional models, if required
                args.kmer_list = kmers_no_models
                rpp(args)
                # add newly generated models to available model list
                for km in kmers_no_models:
                    if os.path.exists(f'{args.out_dir}nns/{km}/nn.h5'):  # check to filter out failed models due to lacking positive examples
                        available_mod_kmers[km] = f'{args.out_dir}nns/{km}'
                    else:
                        warnings.warn(f'model generation failed for {km}, see {args.out_dir}logs. Continuing compilation without it.')
                        kmers.remove(km)
        else:  # Return all k-mers
            kmers = set()
            for _, kmers_in_seq in kmer_candidates_dict.items():
                # Prevent duplicate k-mers
                kmers.update(kmers_in_seq)
            # kmers = get_kmer_candidates_16S(kmer_candidates_dict, args.min_nb_models, 0.0001, filter_list=available_mod_kmers)
            if not len(kmers):
                raise ValueError('Sequences do not contain any of available models!')

    with open(args.parameter_file, 'r') as fh:
        param_dict = yaml.load(fh, yaml.FullLoader)

    target_kmer_dict = {}
    for kmer in kmers:
        model_path = Path(available_mod_kmers[kmer])
        if kmer not in available_mod_kmers or not (model_path / 'nn.h5').exists():
            print(f'No model for {kmer}')
            continue
        target_kmer_dict[kmer] = model_path

    print(f'Saving {len(target_kmer_dict)} kmers for {args.target_16S}')
    mod = compile_model(target_kmer_dict,
                        param_dict['filter_width'], param_dict['filter_stride'],
                        param_dict['threshold'], param_dict['k_frac'])
    mod.save(args.out_model)
    with h5py.File(args.out_model, 'r+') as fh:
        fh.attrs['model_type'] = args.model_type
        fh.attrs['kmer_list'] = ','.join(list(target_kmer_dict))

