from pathlib import Path
from inference.InferenceModel import InferenceModel
import h5py
from db_building.TrainingRead import Read
import argparse
import pandas as pd
import tensorflow as tf


def scan_one_file(file, model, stride=125, verbose=True):
    """Scan a single fast5 file for a target species

    :param file: path to fast5 file
    :param model: Loaded inference model for target species
    :type model: InferenceModel
    :param stride: Stride into which to cut the read
    :param verbose: If true print file and model output
    :return: file name and 1 if species found, otherwise 0
    """
    with h5py.File(file, 'r') as f:
        split_read = Read(f, 'median').get_split_raw_read(250, stride)
    species_found = model.detect_species(split_read)
    if verbose:
        print(f'Processing {file.name}:\t {species_found}')
    return file.name, species_found


def main(args):
    gpus = tf.config.experimental.list_physical_devices('GPU')
    # if args.mem > 0:
    #     tf.config.experimental.set_virtual_device_configuration(gpus[0], [
    #         tf.config.experimental.VirtualDeviceConfiguration(memory_limit=args.mem)])

    batch_size = 64
    # num_files = 67098
    max_files = 67098
    assert args.model_path.suffix == '.h5'
    mod = tf.keras.models.load_model(args.model_path)
    with h5py.File(args.model_path, 'r') as fh:
        model_type = fh.attrs.get('model_type')
        kmer_list = fh.attrs.get('kmer_list').split(',')

    gt = pd.read_csv("/lustre/BIF/nobackup/noord087/HoiCarlos/16Sreads_mockcommunity/ground_truth_with_read_id.csv")
    file_paths = list(args.reads_path.iterdir())

    # Keep only the files that are in the current directory
    gt = gt[gt['file name'].isin([f.name for f in file_paths])]
    # Keep only species that are above 100 reads
    gt = gt.groupby('species id').filter(lambda x: len(x) > 100)

    valid_samples = gt.groupby('species id').sample(100)['file name'].tolist()

    valid_paths = [args.reads_path / f for f in valid_samples]

    raw_in_list = []
    out_list = []
    file_names = []

    for i, file in enumerate(valid_paths, 1):
        # print('Processing', file.name)
        # if file.name not in valid_samples:
        #     continue
        file_names.append(file.name)
        with h5py.File(file, 'r') as fh:
            raw_in_list.append(Read(fh, 'median').raw)
        if len(raw_in_list) % batch_size == 0 or i == len(valid_paths):
            raw = tf.expand_dims(tf.ragged.constant(raw_in_list), -1)
            print(f'Predicted {len(file_names)} files')
            pred = mod(raw).numpy().tolist()
            out_list.extend(pred)
            raw_in_list = []
            if i > max_files:
                break
    d = {'file_name': file_names, 'species_is_found': out_list}
    df = pd.DataFrame(data=d)
    print('Complete')
    df.to_csv(args.out_path, index=False)


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description="""Plot accuracy of
            benchmarked algorithms. Provide only performance CSV and out-dir
            or all arguments except for input-performance-csv""")
    parser.add_argument('--model-path',
                        help='Path to compiled model',
                        required=True, type=Path)
    parser.add_argument('--reads-path',
                        help='Path to directory that contains fast5 reads',
                        required=True, type=Path)
    parser.add_argument('--out-path',
                        help='File path to output txt',
                        required=True, type=Path)
    args = parser.parse_args()
    main(args)
